@@ -25,6 +25,33 @@
 %%----------------------------------------------------------------------
 
 -module(ssh_connection).
+-moduledoc """
+This module provides API functions to send SSH Connection Protocol events to the
+other side of an SSH channel.
+
+The [SSH Connection Protocol (RFC 4254)](http://www.ietf.org/rfc/rfc4254.txt) is used by
+clients and servers, that is, SSH channels, to communicate over the SSH
+connection. The API functions in this module send SSH Connection Protocol
+events, which are received as messages by the remote channel handling the remote
+channel. The Erlang format of thoose messages is (see also
+[below](`t:event/0`)):
+
+`{ssh_cm,` `t:ssh:connection_ref/0` `,` `t:channel_msg/0` `}`
+
+If the `m:ssh_client_channel` behavior is used to implement the channel process,
+these messages are handled by
+[handle_ssh_msg/2](`c:ssh_client_channel:handle_ssh_msg/2`).
+""".
+-moduledoc(#{titles =>
+                 [{type,<<"General">>},
+                  {type,<<"Data Transfer">>},
+                  {type,<<"Closing a Channel">>},
+                  {type,<<"Pseudo-Terminal">>},
+                  {type,<<"Environment Variable">>},
+                  {type,<<"Shell or Command">>},
+                  {type,<<"Window Change">>},
+                  {type,<<"Signals">>},
+                  {type,<<"Exit Status">>}]}).
 
 -include_lib("kernel/include/logger.hrl").
 
@@ -76,11 +103,34 @@
 -type connection_ref() :: ssh:connection_ref().
 -type channel_id()     :: ssh:channel_id().
 
+-doc """
+The status of a request. Corresponds to the `SSH_MSG_CHANNEL_SUCCESS` and
+`SSH_MSG_CHANNEL_FAILURE` values in
+[RFC 4254, Section 5.4](https://tools.ietf.org/html/rfc4254#section-5.4).
+""".
 -type req_status() :: success | failure .
+-doc """
+The result of a call.
+
+If the request reached the peer, was handled and the response reached the
+requesting node the `t:req_status/0` is the status reported from the peer.
+
+If not, the `t:reason/0` indicates what went wrong:
+
+- **`closed`** - indicates that the channel or connection was closed when trying
+  to send the request
+
+- **`timeout`** - indicates that the operation exceeded a time limit
+""".
 -type reason() :: closed | timeout .
 
+-doc(#{}).
 -type result() :: req_status() | {error, reason()} .
 
+-doc """
+The valid values are `0` ("normal") and `1` ("stderr"), see
+[RFC 4254, Section 5.2](https://tools.ietf.org/html/rfc4254#page-8).
+""".
 -type ssh_data_type_code() :: non_neg_integer(). % Only 0 and 1 are used
 
 
@@ -102,7 +152,16 @@
               exec_ch_msg/0
              ]).
 
+-doc(#{title => <<"General">>}).
 -type event() :: {ssh_cm, ssh:connection_ref(), channel_msg()}.
+-doc """
+As mentioned in the introduction, the
+[SSH Connection Protocol](https://tools.ietf.org/html/rfc4254) events are
+handled as messages. When writing a channel handling process without using the
+support by the `m:ssh_client_channel` behavior the process must handle thoose
+messages.
+""".
+-doc(#{title => <<"General">>}).
 -type channel_msg() ::  data_ch_msg()
                       | eof_ch_msg()
                       | closed_ch_msg()
@@ -116,37 +175,95 @@
                       | exit_signal_ch_msg()
                         .
 
+-doc """
+Messages that include a `WantReply` expect the channel handling process to call
+[ssh_connection:reply_request/4](`reply_request/4`) with the boolean value of
+`WantReply` as the second argument.
+""".
+-doc(#{title => <<"General">>}).
 -type want_reply() :: boolean().
 
+-doc """
+Data has arrived on the channel. This event is sent as a result of calling
+[ssh_connection:send/3,4,5](`send/3`).
+""".
+-doc(#{title => <<"Data Transfer">>,
+       rfc => ~"RFC 4254, section 5.2"}).
 -type data_ch_msg() :: {data,
                         ssh:channel_id(),
                         ssh_data_type_code(),
                         Data :: binary()
                        } .
+-doc """
+Indicates that the other side sends no more data. This event is sent as a result
+of calling [ssh_connection:send_eof/2](`send_eof/2`).
+""".
+-doc(#{title => <<"Closing a Channel">>,
+       rfc => ~"RFC 4254, section 5.3"}).
 -type eof_ch_msg() :: {eof,
                        ssh:channel_id()
                       } .
+-doc """
+A signal can be delivered to the remote process/service using the following
+message. Some systems do not support signals, in which case they are to ignore
+this message. There is currently no function to generate this event as the
+signals referred to are on OS-level and not something generated by an Erlang
+program.
+""".
+-doc(#{title => <<"Signals">>,
+       rfc => ~"RFC 4254, section 6.9"}).
 -type signal_ch_msg() :: {signal,
                           ssh:channel_id(),
                           SignalName :: string()
                          } .
+-doc """
+A remote execution can terminate violently because of a signal. Then this
+message can be received. For details on valid string values, see
+[RFC 4254](https://tools.ietf.org/html/rfc4254#section-6.10) Section 6.10, which
+shows a special case of these signals.
+""".
+-doc(#{title => <<"Exit Status">>,
+       rfc => ~"RFC 4254, section 6.10"}).
 -type exit_signal_ch_msg() :: {exit_signal, ssh:channel_id(),
                                ExitSignal :: string(),
                                ErrorMsg :: string(),
                                LanguageString :: string()} .
+-doc """
+When the command running at the other end terminates, the following message can
+be sent to return the exit status of the command. A zero `exit_status` usually
+means that the command terminated successfully. This event is sent as a result
+of calling [ssh_connection:exit_status/3](`exit_status/3`).
+""".
+-doc(#{title => <<"Exit Status">>,
+       rfc => ~"RFC 4254, section 6.10"}).
 -type exit_status_ch_msg() :: {exit_status,
                                ssh:channel_id(),
                                ExitStatus :: non_neg_integer()
                               } .
+-doc """
+This event is sent as a result of calling [ssh_connection:close/2](`close/2`).
+Both the handling of this event and sending it are taken care of by the
+`m:ssh_client_channel` behavior.
+""".
+-doc(#{title => <<"Closing a Channel">>,
+       rfc => ~"RFC 4254, section 5.3"}).
 -type closed_ch_msg() :: {closed,
                           ssh:channel_id()
                          } .
+-doc """
+Environment variables can be passed to the shell/command to be started later.
+This event is sent as a result of calling [ssh_connection:setenv/5](`setenv/5`).
+""".
+-doc(#{title => <<"Environment Variable">>,
+       rfc => ~"RFC 4254, section 6.4"}).
 -type env_ch_msg() :: {env,
                        ssh:channel_id(),
                        want_reply(),
                        Var :: string(),
                        Value :: string()
                       } .
+-doc(#{title => <<"Pseudo-Terminal">>,
+       rfc => ~"RFC 4254, section 6.2"}).
 -type pty_ch_msg() :: {pty,
                        ssh:channel_id(),
                        want_reply(),
@@ -159,13 +276,40 @@
                        }
                       } .
 
+-doc """
+A pseudo-terminal has been requested for the session. `Terminal` is the value of
+the TERM environment variable value, that is, `vt100`. Zero dimension parameters
+must be ignored. The character/row dimensions override the pixel dimensions
+(when non-zero). Pixel dimensions refer to the drawable area of the window.
+`Opcode` in the `TerminalModes` list is the mnemonic name, represented as a
+lowercase Erlang atom, defined in
+[RFC 4254](https://tools.ietf.org/html/rfc4254#section/8), Section 8. It can
+also be an `Opcode` if the mnemonic name is not listed in the RFC. Example:
+`OP code: 53, mnemonic name ECHO erlang atom: echo`. This event is sent as a
+result of calling [ssh_connection:ptty_alloc/4](`ptty_alloc/4`).
+""".
+-doc(#{title => <<"Pseudo-Terminal">>,
+       rfc => ~"RFC 4254, section 6.2"}).
 -type term_mode() :: {Opcode :: atom() | byte(),
                       Value :: non_neg_integer()} .
 
+-doc """
+This message requests that the user default shell is started at the other end.
+This event is sent as a result of calling [ssh_connection:shell/2](`shell/2`).
+""".
+-doc(#{title => <<"Shell or Command">>,
+       rfc => ~"RFC 4254, section 6.5"}).
 -type shell_ch_msg() :: {shell,
                          ssh:channel_id(),
                          want_reply()
                         } .
+-doc """
+When the window (terminal) size changes on the client side, it _can_ send a
+message to the server side to inform it of the new dimensions. No API function
+generates this event.
+""".
+-doc(#{title => <<"Window Change">>,
+       rfc => ~"RFC 4254, section 6.7"}).
 -type window_change_ch_msg() :: {window_change,
                                  ssh:channel_id(),
                                  CharWidth :: non_neg_integer(),
@@ -173,6 +317,12 @@
                                  PixelWidth :: non_neg_integer(),
                                  PixelHeight :: non_neg_integer()
                                 } .
+-doc """
+This message requests that the server starts execution of the given command.
+This event is sent as a result of calling [ssh_connection:exec/4 ](`exec/4`).
+""".
+-doc(#{title => <<"Shell or Command">>,
+       rfc => ~"RFC 4254, section 6.5"}).
 -type exec_ch_msg() :: {exec,
                         ssh:channel_id(),
                         want_reply(),
@@ -182,6 +332,7 @@
 %%% This function is solely to convince all
 %%% checks that the type event() exists...
 -export([dummy/1]).
+-doc false.
 -spec dummy(event()) -> false.
 dummy(_) -> false.
 
@@ -195,15 +346,20 @@ dummy(_) -> false.
 %% application, a system command, or some built-in subsystem.
 %% --------------------------------------------------------------------
 
+-doc(#{equiv => session_channel/4}).
 -spec session_channel(ConnectionRef, Timeout) -> Result when
       ConnectionRef :: ssh:connection_ref(),
       Timeout :: timeout(),
       Result :: {ok, ssh:channel_id()} | {error, reason()} .
 
-session_channel(ConnectionHandler, Timeout) ->
-    session_channel(ConnectionHandler, undefined, undefined, Timeout).
+session_channel(ConnectionRef, Timeout) ->
+    session_channel(ConnectionRef, undefined, undefined, Timeout).
 
 
+-doc """
+Opens a channel for an SSH session. The channel id returned from this function
+is the id used as input to the other functions in this module.
+""".
 -spec session_channel(ConnectionRef, InitialWindowSize, MaxPacketSize, Timeout) -> Result when
       ConnectionRef :: ssh:connection_ref(),
       InitialWindowSize :: pos_integer() | undefined,
@@ -211,8 +367,8 @@ session_channel(ConnectionHandler, Timeout) ->
       Timeout :: timeout(),
       Result :: {ok, ssh:channel_id()} | {error, reason()} .
 
-session_channel(ConnectionHandler, InitialWindowSize, MaxPacketSize, Timeout) ->
-    open_channel(ConnectionHandler, "session", <<>>,
+session_channel(ConnectionRef, InitialWindowSize, MaxPacketSize, Timeout) ->
+    open_channel(ConnectionRef, "session", <<>>,
                  InitialWindowSize,
                  MaxPacketSize,
                  Timeout).
@@ -220,11 +376,12 @@ session_channel(ConnectionHandler, InitialWindowSize, MaxPacketSize, Timeout) ->
 %%--------------------------------------------------------------------
 %% Description: Opens a channel for the given type.
 %% --------------------------------------------------------------------
-open_channel(ConnectionHandler, Type, ChanData, Timeout) ->
-    open_channel(ConnectionHandler, Type, ChanData, undefined, undefined, Timeout).
+-doc false.
+open_channel(ConnectionRef, Type, ChanData, Timeout) ->
+    open_channel(ConnectionRef, Type, ChanData, undefined, undefined, Timeout).
 
-open_channel(ConnectionHandler, Type, ChanData, InitialWindowSize, MaxPacketSize, Timeout) ->
-    case ssh_connection_handler:open_channel(ConnectionHandler, Type, ChanData,
+open_channel(ConnectionRef, Type, ChanData, InitialWindowSize, MaxPacketSize, Timeout) ->
+    case ssh_connection_handler:open_channel(ConnectionRef, Type, ChanData,
                                              InitialWindowSize, MaxPacketSize,
                                              Timeout) of
         {open, Channel} ->
@@ -237,14 +394,40 @@ open_channel(ConnectionHandler, Type, ChanData, InitialWindowSize, MaxPacketSize
 %% Description: Will request that the server start the
 %% execution of the given command. 
 %%--------------------------------------------------------------------
+-doc """
+Is to be called by a client-channel process to request that the server starts
+executing the given command. The result is several messages according to the
+following pattern. The last message is a channel close message, as the `exec`
+request is a one-time execution that closes the channel when it is done.
+
+- **N x [data message(s)](`t:data_ch_msg/0`)** - The result of executing the
+  command can be only one line or thousands of lines depending on the command.
+
+- **0 or 1 x [eof message](`t:eof_ch_msg/0`)** - Indicates that no more data is
+  to be sent.
+
+- **0 or 1 x [exit signal message](`t:exit_signal_ch_msg/0`)** - Not all systems
+  send signals. For details on valid string values, see RFC 4254, Section 6.10
+
+- **0 or 1 x [exit status message](`t:exit_status_ch_msg/0`)** - It is
+  recommended by the SSH Connection Protocol to send this message, but that is
+  not always the case.
+
+- **1 x [closed status message](`t:closed_ch_msg/0`)** - Indicates that the
+  `ssh_client_channel` started for the execution of the command has now been
+  shut down.
+
+See the User's Guide section on
+[One-Time Execution](using_ssh.md#one-time-execution) for examples.
+""".
 -spec exec(ConnectionRef, ChannelId, Command, Timeout) -> result() when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
       Command :: string(),
       Timeout :: timeout().
 
-exec(ConnectionHandler, ChannelId, Command, TimeOut) ->
-    ssh_connection_handler:request(ConnectionHandler, self(), ChannelId, "exec",
+exec(ConnectionRef, ChannelId, Command, TimeOut) ->
+    ssh_connection_handler:request(ConnectionRef, self(), ChannelId, "exec",
 				   true, [?string(Command)], TimeOut).
 
 %%--------------------------------------------------------------------
@@ -252,57 +435,97 @@ exec(ConnectionHandler, ChannelId, Command, TimeOut) ->
 %% defined in /etc/passwd in UNIX systems) be started at the other
 %% end.
 %%--------------------------------------------------------------------
+-doc """
+Is to be called by a client channel process to request that the user default
+shell (typically defined in /etc/passwd in Unix systems) is executed at the
+server end.
+
+Note: the return value is `ok` instead of `success` unlike in other functions in
+this module. This is a fault that was introduced so long ago that any change
+would break a large number of existing software.
+""".
 -spec shell(ConnectionRef, ChannelId) -> Result when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
       Result :: ok | success | failure | {error, timeout} .
 
-shell(ConnectionHandler, ChannelId) ->
-    ssh_connection_handler:request(ConnectionHandler, self(), ChannelId,
+shell(ConnectionRef, ChannelId) ->
+    ssh_connection_handler:request(ConnectionRef, self(), ChannelId,
  				   "shell", false, <<>>, 0).
 %%--------------------------------------------------------------------
 %%
 %% Description: Executes a predefined subsystem.
 %%--------------------------------------------------------------------
+-doc """
+Is to be called by a client-channel process for requesting to execute a
+predefined subsystem on the server.
+
+The function [`subsystem/4`](`subsystem/4`) and subsequent calls of
+[send/3,4,5](`send/3`) must be executed in the same process.
+""".
 -spec subsystem(ConnectionRef, ChannelId, Subsystem, Timeout) -> result() when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
       Subsystem  :: string(),
       Timeout :: timeout().
 
-subsystem(ConnectionHandler, ChannelId, SubSystem, TimeOut) ->
-     ssh_connection_handler:request(ConnectionHandler, self(),
+subsystem(ConnectionRef, ChannelId, SubSystem, TimeOut) ->
+     ssh_connection_handler:request(ConnectionRef, self(),
 				    ChannelId, "subsystem", 
 				    true, [?string(SubSystem)], TimeOut).
 %%--------------------------------------------------------------------
 %% Description: Sends channel data.
 %%--------------------------------------------------------------------
+-doc(#{equiv => send/5}).
 -spec send(connection_ref(), channel_id(), iodata()) ->
-		  ok | {error, timeout | closed}.
+		  ok | {error, reason()}.
 
-send(ConnectionHandler, ChannelId, Data) ->
-    send(ConnectionHandler, ChannelId, 0, Data, infinity).
+send(ConnectionRef, ChannelId, Data) ->
+    send(ConnectionRef, ChannelId, 0, Data, infinity).
 
 
--spec send(connection_ref(), channel_id(), iodata(), timeout()) -> ok |  {error, reason()};
-          (connection_ref(), channel_id(), ssh_data_type_code(), iodata()) -> ok |  {error, reason()}.
+-doc """
+send(ConnectionRef, ChannelId, Type, Data)
 
-send(ConnectionHandler, ChannelId, Data, TimeOut) when is_integer(TimeOut) ->
-    send(ConnectionHandler, ChannelId, 0, Data, TimeOut);
+Depending on input arguments equivalent to one of `send/5` calls specified below.
 
-send(ConnectionHandler, ChannelId, Data, infinity) ->
-    send(ConnectionHandler, ChannelId, 0, Data, infinity);
+Equivalent to [send(ConnectionRef, ChannelId, 0, Data, TimeOut)](`send/5`) if
+called with TimeOut being integer.
 
-send(ConnectionHandler, ChannelId, Type, Data) ->
-    send(ConnectionHandler, ChannelId, Type, Data, infinity).
+Equivalent to [send(ConnectionRef, ChannelId, 0, Data, infinity)](`send/5`) if
+called with TimeOut being infinity atom.
+
+Equivalent to [send(ConnectionRef, ChannelId, Type, Data, infinity)](`send/5`) if
+called with last argument which is not integer or infinity atom.
+""".
+
+-spec send(connection_ref(), channel_id(), iodata(), timeout()) -> ok | {error, reason()};
+          (connection_ref(), channel_id(), ssh_data_type_code(), iodata()) -> ok | {error, reason()}.
+
+send(ConnectionRef, ChannelId, Data, TimeOut) when is_integer(TimeOut) ->
+    send(ConnectionRef, ChannelId, 0, Data, TimeOut);
+
+send(ConnectionRef, ChannelId, Data, infinity) ->
+    send(ConnectionRef, ChannelId, 0, Data, infinity);
+
+send(ConnectionRef, ChannelId, Type, Data) ->
+    send(ConnectionRef, ChannelId, Type, Data, infinity).
 
 
--spec send(connection_ref(), channel_id(), ssh_data_type_code(), iodata(), timeout()) -> ok |  {error, reason()}.
+-doc """
+Is to be called by client- and server-channel processes to send data to each
+other.
 
-send(ConnectionHandler, ChannelId, Type, Data, TimeOut) ->
-    ssh_connection_handler:send(ConnectionHandler, ChannelId,
+The function `subsystem/4` and subsequent calls of `send/3,4,5` must be executed
+in the same process.
+""".
+-spec send(connection_ref(), channel_id(), ssh_data_type_code(), iodata(), timeout()) -> ok | {error, reason()}.
+
+send(ConnectionRef, ChannelId, Type, Data, TimeOut) ->
+    ssh_connection_handler:send(ConnectionRef, ChannelId,
 				Type, Data, TimeOut).
 %%--------------------------------------------------------------------
+-doc "Sends EOF on channel `ChannelId`.".
 -spec send_eof(ConnectionRef, ChannelId) -> ok  | {error, closed} when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id().
@@ -310,10 +533,22 @@ send(ConnectionHandler, ChannelId, Type, Data, TimeOut) ->
 %%
 %% Description: Sends eof on the channel <ChannelId>.
 %%--------------------------------------------------------------------
-send_eof(ConnectionHandler, Channel) ->
-    ssh_connection_handler:send_eof(ConnectionHandler, Channel).
+send_eof(ConnectionRef, Channel) ->
+    ssh_connection_handler:send_eof(ConnectionRef, Channel).
 
 %%--------------------------------------------------------------------
+-doc """
+Adjusts the SSH flow control window. This is to be done by both the client- and
+server-side channel processes.
+
+> #### Note {: .info }
+>
+> Channels implemented with the `m:ssh_client_channel` behavior do not normally
+> need to call this function as flow control is handled by the behavior. The
+> behavior adjusts the window every time the callback
+> [handle_ssh_msg/2](`c:ssh_client_channel:handle_ssh_msg/2`) returns after
+> processing channel data.
+""".
 -spec adjust_window(ConnectionRef, ChannelId, NumOfBytes) -> ok when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
@@ -322,10 +557,14 @@ send_eof(ConnectionHandler, Channel) ->
 %%
 %% Description: Adjusts the ssh flowcontrol window.
 %%--------------------------------------------------------------------
-adjust_window(ConnectionHandler, Channel, Bytes) ->
-    ssh_connection_handler:adjust_window(ConnectionHandler, Channel, Bytes).
+adjust_window(ConnectionRef, Channel, Bytes) ->
+    ssh_connection_handler:adjust_window(ConnectionRef, Channel, Bytes).
 
 %%--------------------------------------------------------------------
+-doc """
+Environment variables can be passed before starting the shell/command. Is to be
+called by a client channel processes.
+""".
 -spec setenv(ConnectionRef, ChannelId, Var, Value, Timeout) -> success when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
@@ -336,11 +575,11 @@ adjust_window(ConnectionHandler, Channel, Bytes) ->
 %%
 %% Description: Environment variables may be passed to the shell/command to be
 %% started later.
-setenv(ConnectionHandler, ChannelId, Var, Value, TimeOut) ->
-    setenv(ConnectionHandler, ChannelId, true, Var, Value, TimeOut).
+setenv(ConnectionRef, ChannelId, Var, Value, TimeOut) ->
+    setenv(ConnectionRef, ChannelId, true, Var, Value, TimeOut).
 
-setenv(ConnectionHandler, ChannelId, WantReply, Var, Value, TimeOut) ->
-    case ssh_connection_handler:request(ConnectionHandler, ChannelId,
+setenv(ConnectionRef, ChannelId, WantReply, Var, Value, TimeOut) ->
+    case ssh_connection_handler:request(ConnectionRef, ChannelId,
                                         "env", WantReply,
                                         [?string(Var), ?string(Value)], TimeOut) of
         ok when WantReply == false ->
@@ -349,6 +588,16 @@ setenv(ConnectionHandler, ChannelId, WantReply, Var, Value, TimeOut) ->
             Reply
     end.
 %%--------------------------------------------------------------------
+-doc """
+A server- or client-channel process can choose to close their session by sending
+a close event.
+
+> #### Note {: .info }
+>
+> This function is called by the `ssh_client_channel` behavior when the channel
+> is terminated, see `m:ssh_client_channel`. Thus, channels implemented with the
+> behavior are not to call this function explicitly.
+""".
 -spec close(ConnectionRef, ChannelId) -> ok when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id().
@@ -356,10 +605,16 @@ setenv(ConnectionHandler, ChannelId, WantReply, Var, Value, TimeOut) ->
 %%
 %% Description: Sends a close message on the channel <ChannelId>.
 %%--------------------------------------------------------------------
-close(ConnectionHandler, ChannelId) ->
-    ssh_connection_handler:close(ConnectionHandler, ChannelId).
+close(ConnectionRef, ChannelId) ->
+    ssh_connection_handler:close(ConnectionRef, ChannelId).
 
 %%--------------------------------------------------------------------
+-doc """
+Sends status replies to requests where the requester has stated that it wants a
+status report, that is, `WantReply = true`. If `WantReply` is `false`, calling
+this function becomes a "noop". Is to be called while handling an SSH Connection
+Protocol message containing a `WantReply` boolean value.
+""".
 -spec reply_request(ConnectionRef, WantReply, Status, ChannelId) -> ok when
       ConnectionRef :: ssh:connection_ref(),
       WantReply :: boolean(),
@@ -369,34 +624,58 @@ close(ConnectionHandler, ChannelId) ->
 %%
 %% Description: Send status replies to requests that want such replies.
 %%--------------------------------------------------------------------
-reply_request(ConnectionHandler, true, Status, ChannelId) ->
-    ssh_connection_handler:reply_request(ConnectionHandler, Status, ChannelId);
+reply_request(ConnectionRef, true, Status, ChannelId) ->
+    ssh_connection_handler:reply_request(ConnectionRef, Status, ChannelId);
 reply_request(_,false, _, _) ->
     ok.
 
 %%--------------------------------------------------------------------
 %% Description: Sends a ssh connection protocol pty_req.
 %%--------------------------------------------------------------------
+-doc(#{equiv => ptty_alloc/4}).
+-doc(#{since => <<"OTP 17.5">>}).
 -spec ptty_alloc(ConnectionRef, ChannelId, Options) -> result() when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
       Options  :: proplists:proplist().
 
-ptty_alloc(ConnectionHandler, Channel, Options) ->
-    ptty_alloc(ConnectionHandler, Channel, Options, infinity).
+ptty_alloc(ConnectionRef, Channel, Options) ->
+    ptty_alloc(ConnectionRef, Channel, Options, infinity).
 
 
+-doc """
+Sends an SSH Connection Protocol `pty_req`, to allocate a pseudo-terminal. Is to
+be called by an SSH client process.
+
+Options:
+
+- **\{term, string()\}** - Defaults to _os:getenv("TERM")_ or _vt100_ if it is
+  undefined.
+
+- **\{width, integer()\}** - Defaults to 80 if `pixel_width` is not defined.
+
+- **\{height, integer()\}** - Defaults to 24 if `pixel_height` is not defined.
+
+- **\{pixel_width, integer()\}** - Is disregarded if `width` is defined.
+
+- **\{pixel_height, integer()\}** - Is disregarded if `height` is defined.
+
+- **\{pty_opts, \[\{posix_atom(), integer()\}]\}** - Option can be an empty
+  list. Otherwise, see possible _POSIX_ names in Section 8 in
+  [RFC 4254](http://www.ietf.org/rfc/rfc4254.txt).
+""".
+-doc(#{since => <<"OTP 17.4">>}).
 -spec ptty_alloc(ConnectionRef, ChannelId, Options, Timeout) -> result() when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
       Options  :: proplists:proplist(),
       Timeout :: timeout().
 
-ptty_alloc(ConnectionHandler, Channel, Options0, TimeOut) ->
+ptty_alloc(ConnectionRef, Channel, Options0, TimeOut) ->
     TermData = backwards_compatible(Options0, []), % FIXME
     {Width, PixWidth} = pty_default_dimensions(width, TermData),
     {Height, PixHeight} = pty_default_dimensions(height, TermData),
-    pty_req(ConnectionHandler, Channel,
+    pty_req(ConnectionRef, Channel,
 	    proplists:get_value(term, TermData, os:getenv("TERM", ?DEFAULT_TERMINAL)),
 	    proplists:get_value(width, TermData, Width),
 	    proplists:get_value(height, TermData, Height),
@@ -410,26 +689,33 @@ ptty_alloc(ConnectionHandler, Channel, Options0, TimeOut) ->
 %% initial contributed ssh application. They are untested. Do we want them?
 %% Should they be documented and tested?
 %%--------------------------------------------------------------------
-window_change(ConnectionHandler, Channel, Width, Height) ->
-    window_change(ConnectionHandler, Channel, Width, Height, 0, 0).
-window_change(ConnectionHandler, Channel, Width, Height,
+-doc false.
+window_change(ConnectionRef, Channel, Width, Height) ->
+    window_change(ConnectionRef, Channel, Width, Height, 0, 0).
+-doc false.
+window_change(ConnectionRef, Channel, Width, Height,
 	      PixWidth, PixHeight) ->
-    ssh_connection_handler:request(ConnectionHandler, Channel,
+    ssh_connection_handler:request(ConnectionRef, Channel,
 				   "window-change", false, 
 				   [?uint32(Width), ?uint32(Height),
 				    ?uint32(PixWidth), ?uint32(PixHeight)], 0).
 
-signal(ConnectionHandler, Channel, Sig) ->
-    ssh_connection_handler:request(ConnectionHandler, Channel,
+-doc false.
+signal(ConnectionRef, Channel, Sig) ->
+    ssh_connection_handler:request(ConnectionRef, Channel,
 				   "signal", false, [?string(Sig)], 0).
 
 
+-doc """
+Is to be called by a server-channel process to send the exit status of a command
+to the client.
+""".
 -spec exit_status(ConnectionRef, ChannelId, Status) -> ok when
       ConnectionRef :: ssh:connection_ref(),
       ChannelId :: ssh:channel_id(),
       Status  :: integer().
-exit_status(ConnectionHandler, Channel, Status) ->
-    ssh_connection_handler:request(ConnectionHandler, Channel,
+exit_status(ConnectionRef, Channel, Status) ->
+    ssh_connection_handler:request(ConnectionRef, Channel,
 				   "exit-status", false, [?uint32(Status)], 0).
 
 %%--------------------------------------------------------------------
@@ -441,6 +727,7 @@ exit_status(ConnectionHandler, Channel, Status) ->
 %%% ssh_connection:send (executed in the ssh_connection_state machine)
 %%%
 
+-doc false.
 channel_data(ChannelId, DataType, Data0, 
 	     #connection{channel_cache = Cache} = Connection,
 	     From) ->
@@ -470,6 +757,7 @@ channel_data(ChannelId, DataType, Data0,
 %%% Replies {Reply, UpdatedConnection}
 %%%
 
+-doc false.
 handle_msg(#ssh_msg_disconnect{code = Code, description = Description}, Connection, _, _SSH) ->
     {disconnect, {Code, Description}, handle_stop(Connection)};
 
@@ -481,9 +769,10 @@ handle_msg(Msg, Connection, server, Ssh = #ssh{authenticated = false}) ->
     %% respond by disconnecting, preferably with a proper disconnect message
     %% sent to ease troubleshooting.
     MsgFun = fun(M) ->
+                     MaxLogItemLen = ?GET_OPT(max_log_item_len, Ssh#ssh.opts),
                      io_lib:format("Connection terminated. Unexpected message for unauthenticated user."
                                    " Message:  ~w", [M],
-                                   [{chars_limit, ssh_lib:max_log_len(Ssh)}])
+                                   [{chars_limit, MaxLogItemLen}])
              end,
     ?LOG_DEBUG(MsgFun, [Msg]),
     {disconnect, {?SSH_DISCONNECT_PROTOCOL_ERROR, "Connection refused"}, handle_stop(Connection)};
@@ -1022,6 +1311,7 @@ handle_msg(#ssh_msg_request_success{data = Data},
 %%% Returns pending responses to be delivered to the peer when a
 %%% Channel/Connection closes
 %%%
+-doc false.
 handle_stop(#connection{channel_cache = Cache} = Connection0) ->
     {Connection, Replies} = 
 	ssh_client_channel:cache_foldl(
@@ -1038,13 +1328,16 @@ handle_stop(#connection{channel_cache = Cache} = Connection0) ->
 %%% channel_*_msg(...)
 %%% Returns a #ssh_msg_....{} for channel operations.
 %%%
+-doc false.
 channel_adjust_window_msg(ChannelId, Bytes) ->
     #ssh_msg_channel_window_adjust{recipient_channel = ChannelId,
 				   bytes_to_add = Bytes}.
 
+-doc false.
 channel_close_msg(ChannelId) ->
     #ssh_msg_channel_close {recipient_channel = ChannelId}.
 
+-doc false.
 channel_data_msg(ChannelId, 0, Data) ->
     #ssh_msg_channel_data{recipient_channel = ChannelId,
 			  data = Data};
@@ -1053,12 +1346,15 @@ channel_data_msg(ChannelId, Type, Data) ->
 				    data_type_code = Type,
 				    data = Data}.
 
+-doc false.
 channel_eof_msg(ChannelId) ->
     #ssh_msg_channel_eof{recipient_channel = ChannelId}.
 
+-doc false.
 channel_failure_msg(ChannelId) ->
     #ssh_msg_channel_failure{recipient_channel = ChannelId}.
 
+-doc false.
 channel_open_msg(Type, ChannelId, WindowSize, MaxPacketSize, Data) ->
     #ssh_msg_channel_open{channel_type = Type,
 			  sender_channel = ChannelId,
@@ -1067,30 +1363,35 @@ channel_open_msg(Type, ChannelId, WindowSize, MaxPacketSize, Data) ->
 			  data = Data
 			 }.
 
+-doc false.
 channel_open_confirmation_msg(RemoteId, LID, WindowSize, PacketSize) ->
     #ssh_msg_channel_open_confirmation{recipient_channel = RemoteId,
 				       sender_channel = LID,
 				       initial_window_size = WindowSize,
 				       maximum_packet_size = PacketSize}.
 
+-doc false.
 channel_open_failure_msg(RemoteId, Reason, Description, Lang) ->
     #ssh_msg_channel_open_failure{recipient_channel = RemoteId,
 				  reason = Reason,
 				  description = Description,
 				  lang = Lang}.
 
+-doc false.
 channel_status_msg({success, ChannelId}) ->
     channel_success_msg(ChannelId);
 
 channel_status_msg({failure, ChannelId}) ->
     channel_failure_msg(ChannelId).
 
+-doc false.
 channel_request_msg(ChannelId, Type, WantReply, Data) ->
     #ssh_msg_channel_request{recipient_channel = ChannelId,
 			     request_type = Type,
 			     want_reply = WantReply,
 			     data = Data}.
 
+-doc false.
 channel_success_msg(ChannelId) ->
     #ssh_msg_channel_success{recipient_channel = ChannelId}.
 
@@ -1098,20 +1399,24 @@ channel_success_msg(ChannelId) ->
 %%% request_*_msg(...)
 %%% Returns a #ssh_msg_....{}
 %%%
+-doc false.
 request_global_msg(Name, WantReply, Data) ->
     #ssh_msg_global_request{name = Name,
                             want_reply = WantReply,
                             data = Data}.
 
+-doc false.
 request_failure_msg() ->
     #ssh_msg_request_failure{}.
 
+-doc false.
 request_success_msg(Data) ->
     #ssh_msg_request_success{data = Data}.
 
 %%%----------------------------------------------------------------
 %%%
 %%%
+-doc false.
 encode_ip(Addr) when is_tuple(Addr) ->
     case catch inet_parse:ntoa(Addr) of
 	{'EXIT',_} -> false;
@@ -1299,9 +1604,9 @@ flow_control(_,_,_) ->
 %%% Pseudo terminal stuff
 %%% 
 
-pty_req(ConnectionHandler, Channel, Term, Width, Height,
+pty_req(ConnectionRef, Channel, Term, Width, Height,
 	 PixWidth, PixHeight, PtyOpts, TimeOut) ->
-    ssh_connection_handler:request(ConnectionHandler,
+    ssh_connection_handler:request(ConnectionRef,
 				   Channel, "pty-req", true,
 				   [?string(Term),
 				    ?uint32(Width), ?uint32(Height),
@@ -1322,6 +1627,7 @@ pty_default_dimensions(Dimension, TermData) ->
 	    end
     end.
 
+-doc false.
 encode_pty_opts(Opts) ->
     Bin = list_to_binary(encode_pty_opts2(Opts)),
     <<?STRING(Bin)>>.
@@ -1441,6 +1747,7 @@ encode_pty_opts2([{tty_op_ispeed,Value} | Opts]) ->
 encode_pty_opts2([{tty_op_ospeed,Value} | Opts]) ->
     [?TTY_OP_OSPEED, ?uint32(Value) | encode_pty_opts2(Opts)].
 
+-doc false.
 decode_pty_opts(<<>>) ->		     
     [];
 decode_pty_opts(<<0, 0, 0, 0>>) ->
@@ -1607,14 +1914,15 @@ request_reply_or_data(#channel{local_id = ChannelId, user = ChannelPid},
     end.
 
 %%%----------------------------------------------------------------
-send_environment_vars(ConnectionHandler, Channel, VarNames) ->
+-doc false.
+send_environment_vars(ConnectionRef, Channel, VarNames) ->
     lists:foldl(
       fun(Var, success) ->
               case os:getenv(Var) of
                   false ->
                       success;
                   Value ->
-                      setenv(ConnectionHandler, Channel, false,
+                      setenv(ConnectionRef, Channel, false,
                              Var, Value, infinity)
               end
       end, success, VarNames).
