@@ -27,6 +27,7 @@
 %% ----------------------------------------------------------------------
 
 -module(ssh_connection_handler).
+-moduledoc false.
 
 -behaviour(gen_statem).
 
@@ -459,6 +460,8 @@ init_ssh_record(Role, Socket, PeerAddr, Opts) ->
                            PeerName0 when is_list(PeerName0) ->
                                PeerName0
                        end,
+            ssh_lib:set_label(client,
+                              {connection, list_to_binary(PeerName), Socket}),
             S1 =
                 S0#ssh{c_vsn = Vsn,
                        c_version = Version,
@@ -477,6 +480,7 @@ init_ssh_record(Role, Socket, PeerAddr, Opts) ->
                   };
 
 	server ->
+            ssh_lib:set_label(server, {connection, Socket}),
 	    S0#ssh{s_vsn = Vsn,
 		   s_version = Version,
 		   userauth_methods = string:tokens(AuthMethods, ","),
@@ -609,18 +613,17 @@ handle_event(cast, socket_control, {wait_for_socket, Role},
 handle_event(internal, socket_ready, {hello,_}=StateName, #data{ssh_params = Ssh0} = D) ->
     VsnMsg = ssh_transport:hello_version_msg(string_version(Ssh0)),
     send_bytes(VsnMsg, D),
-    case inet:getopts(Socket=D#data.socket, [buffer]) of
-	{ok, [{buffer,Size}]} ->
+    case inet:getopts(Socket=D#data.socket, [recbuf]) of
+	{ok, [{recbuf,Size}]} ->
 	    %% Set the socket to the hello text line handling mode:
 	    inet:setopts(Socket, [{packet, line},
 				  {active, once},
 				  % Expecting the version string which might
 				  % be max ?MAX_PROTO_VERSION bytes:
-				  {buffer, ?MAX_PROTO_VERSION},
-				  {packet_size, ?MAX_PROTO_VERSION},
+				  {recbuf, ?MAX_PROTO_VERSION},
 				  {nodelay,true}]),
             Time = ?GET_OPT(hello_timeout, Ssh0#ssh.opts, infinity),
-	    {keep_state, D#data{inet_initial_buffer_size=Size}, [{state_timeout,Time,no_hello_received}] };
+	    {keep_state, D#data{inet_initial_recbuf_size=Size}, [{state_timeout,Time,no_hello_received}] };
 
 	Other ->
             ?call_disconnectfun_and_log_cond("Option return", 
@@ -649,12 +652,11 @@ handle_event(internal, {version_exchange,Version}, {hello,Role}, D0) ->
     case handle_version(NumVsn, StrVsn, D0#data.ssh_params) of
 	{ok, Ssh1} ->
 	    %% Since the hello part is finished correctly, we set the
-	    %% socket to the packet handling mode (including buffer size):
+	    %% socket to the packet handling mode (including recbuf size):
 	    inet:setopts(D0#data.socket, [{packet,0},
 					 {mode,binary},
 					 {active, once},
-					 {buffer, D0#data.inet_initial_buffer_size},
-					 {packet_size, 0}]),
+					 {recbuf, D0#data.inet_initial_recbuf_size}]),
 	    {KeyInitMsg, SshPacket, Ssh} = ssh_transport:key_exchange_init_msg(Ssh1),
 	    send_bytes(SshPacket, D0),
             D = D0#data{ssh_params = Ssh,
@@ -1184,21 +1186,12 @@ handle_event(info, {Proto, Sock, NewData}, StateName,
                                       {next_event, internal, Msg}
 				    ]}
 	    catch
-		Class:Reason0:Stacktrace  ->
-                    Reason = ssh_lib:trim_reason(Reason0),
-                    MsgFun =
-                        fun(debug) ->
-                                io_lib:format("Bad packet: Decrypted, but can't decode~n~p:~p~n~p",
-                                              [Class,Reason,Stacktrace],
-                                              [{chars_limit, ssh_lib:max_log_len(SshParams)}]);
-                           (_) ->
-                                io_lib:format("Bad packet: Decrypted, but can't decode ~p:~p",
-                                              [Class, Reason],
-                                              [{chars_limit, ssh_lib:max_log_len(SshParams)}])
-                        end,
+		C:E:ST  ->
+                    MaxLogItemLen = ?GET_OPT(max_log_item_len,SshParams#ssh.opts),
                     {Shutdown, D} =
                         ?send_disconnect(?SSH_DISCONNECT_PROTOCOL_ERROR,
-                                         ?SELECT_MSG(MsgFun),
+                                         io_lib:format("Bad packet: Decrypted, but can't decode~n~p:~p~n~p",
+                                                       [C,E,ST], [{chars_limit, MaxLogItemLen}]),
                                          StateName, D1),
                     {stop, Shutdown, D}
 	    end;
@@ -1228,20 +1221,12 @@ handle_event(info, {Proto, Sock, NewData}, StateName,
                                  StateName, D0),
             {stop, Shutdown, D}
     catch
-	Class:Reason0:Stacktrace ->
-            MsgFun =
-                fun(debug) ->
-                        io_lib:format("Bad packet: Couldn't decrypt~n~p:~p~n~p",
-                                      [Class,Reason0,Stacktrace],
-                                      [{chars_limit, ssh_lib:max_log_len(SshParams)}]);
-                   (_) ->
-                        Reason = ssh_lib:trim_reason(Reason0),
-                        io_lib:format("Bad packet: Couldn't decrypt~n~p:~p",
-                                      [Class,Reason],
-                                      [{chars_limit, ssh_lib:max_log_len(SshParams)}])
-                end,
+	C:E:ST ->
+            MaxLogItemLen = ?GET_OPT(max_log_item_len,SshParams#ssh.opts),
             {Shutdown, D} =
-                ?send_disconnect(?SSH_DISCONNECT_PROTOCOL_ERROR, ?SELECT_MSG(MsgFun),
+                ?send_disconnect(?SSH_DISCONNECT_PROTOCOL_ERROR,
+                                 io_lib:format("Bad packet: Couldn't decrypt~n~p:~p~n~p",
+                                               [C,E,ST], [{chars_limit, MaxLogItemLen}]),
                                  StateName, D0),
             {stop, Shutdown, D}
     end;
