@@ -12,24 +12,26 @@
 %%
 %% @copyright 2000-2003 Richard Carlsson
 %% @author Richard Carlsson <carlsson.richard@gmail.com>
-%% @doc Support functions for property lists.
-%%
-%% <p>Property lists are ordinary lists containing entries in the form
-%% of either tuples, whose first elements are keys used for lookup and
-%% insertion, or atoms, which work as shorthand for tuples <code>{Atom,
-%% true}</code>. (Other terms are allowed in the lists, but are ignored
-%% by this module.) If there is more than one entry in a list for a
-%% certain key, the first occurrence normally overrides any later
-%% (irrespective of the arity of the tuples).</p>
-%%
-%% <p>Property lists are useful for representing inherited properties,
-%% such as options passed to a function where a user may specify options
-%% overriding the default settings, object properties, annotations,
-%% etc.</p>
-%%
-%% % @type property() = atom() | tuple()
-
 -module(proplists).
+-moduledoc(#{ authors => [~"Richard Carlsson <carlsson.richard@gmail.com>"] }).
+-moduledoc """
+Support functions for property lists.
+
+Property lists are ordinary lists containing entries in the form of either
+tuples, whose first elements are keys used for lookup and insertion, or atoms,
+which work as shorthand for tuples `{Atom, true}`. (Other terms are allowed in
+the lists, but are ignored by this module.) If there is more than one entry in a
+list for a certain key, the first occurrence normally overrides any later
+(irrespective of the arity of the tuples).
+
+Property lists are useful for representing inherited properties, such as options
+passed to a function where a user can specify options overriding the default
+settings, object properties, annotations, and so on.
+
+Two keys are considered equal if they match (`=:=`). That is, numbers are
+compared literally rather than by value, so that, for example, `1` and `1.0` are
+different keys.
+""".
 
 -export([property/1, property/2, unfold/1, compact/1, lookup/2,
 	 lookup_all/2, is_defined/2, get_value/2, get_value/3,
@@ -42,18 +44,20 @@
 
 -export_type([property/0, proplist/0]).
 
+-doc "A property item within a list".
 -type property()  :: atom() | tuple().
+-doc "A list of `t:property/0`, also knows as a proplist.".
 -type proplist()  :: [property()].
 
 %% ---------------------------------------------------------------------
 
-%% @doc Creates a normal form (minimal) representation of a property. If
-%% <code>PropertyIn</code> is <code>{Key, true}</code> where
-%% <code>Key</code> is an atom, this returns <code>Key</code>, otherwise
-%% the whole term <code>PropertyIn</code> is returned.
-%%
-%% @see property/2
+-doc """
+Creates a normal form (minimal) representation of a property. If `PropertyIn` is
+`{Key, true}`, where `Key` is an atom, `Key` is returned, otherwise the whole
+term `PropertyIn` is returned.
 
+See also `property/2`.
+""".
 -spec property(PropertyIn) -> PropertyOut when
       PropertyIn :: property(),
       PropertyOut :: property().
@@ -64,13 +68,13 @@ property(Property) ->
     Property.
 
 
-%% @doc Creates a normal form (minimal) representation of a simple
-%% key/value property. Returns <code>Key</code> if <code>Value</code> is
-%% <code>true</code> and <code>Key</code> is an atom, otherwise a tuple
-%% <code>{Key, Value}</code> is returned.
-%%
-%% @see property/1
+-doc """
+Creates a normal form (minimal) representation of a simple key/value property.
+Returns `Key` if `Value` is `true` and `Key` is an atom, otherwise a tuple
+`{Key, Value}` is returned.
 
+See also `property/1`.
+""".
 -spec property(Key, Value) -> Property when
       Key :: term(),
       Value :: term(),
@@ -84,11 +88,11 @@ property(Key, Value) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Unfolds all occurrences of atoms in <code>ListIn</code> to tuples
-%% <code>{Atom, true}</code>.
-%%
-%% @see compact/1
+-doc """
+Unfolds all occurrences of atoms in `ListIn` to tuples `{Atom, true}`.
 
+See also `compact/1`.
+""".
 -spec unfold(ListIn) -> ListOut when
       ListIn :: [term()],
       ListOut :: [term()].
@@ -102,12 +106,12 @@ unfold([P | Ps]) ->
 unfold([]) ->
     [].
 
-%% @doc Minimizes the representation of all entries in the list. This is
-%% equivalent to <code>[property(P) || P &lt;- ListIn]</code>.
-%%
-%% @see unfold/1
-%% @see property/1
+-doc """
+Minimizes the representation of all entries in the list. This is equivalent to
+`[property(P) || P <- ListIn]`.
 
+See also `property/1`, `unfold/1`.
+""".
 -spec compact(ListIn) -> ListOut when
       ListIn :: [property()],
       ListOut :: [property()].
@@ -118,15 +122,13 @@ compact(ListIn) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Returns the first entry associated with <code>Key</code> in
-%% <code>List</code>, if one exists, otherwise returns
-%% <code>none</code>. For an atom <code>A</code> in the list, the tuple
-%% <code>{A, true}</code> is the entry associated with <code>A</code>.
-%%
-%% @see lookup_all/2
-%% @see get_value/2
-%% @see get_bool/2
+-doc """
+Returns the first entry associated with `Key` in `List`, if one exists,
+otherwise returns `none`. For an atom `A` in the list, the tuple `{A, true}` is
+the entry associated with `A`.
 
+See also `get_bool/2`, `get_value/2`, `lookup_all/2`.
+""".
 -spec lookup(Key, List) -> 'none' | tuple() when
       Key :: term(),
       List :: [term()].
@@ -143,12 +145,12 @@ lookup(Key, [P | Ps]) ->
 lookup(_Key, []) ->
     none.
 
-%% @doc Returns the list of all entries associated with <code>Key</code>
-%% in <code>List</code>. If no such entry exists, the result is the
-%% empty list.
-%%
-%% @see lookup/2
+-doc """
+Returns the list of all entries associated with `Key` in `List`. If no such
+entry exists, the result is the empty list.
 
+See also `lookup/2`.
+""".
 -spec lookup_all(Key, List) -> [tuple()] when
       Key :: term(),
       List :: [term()].
@@ -167,10 +169,10 @@ lookup_all(_Key, []) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Returns <code>true</code> if <code>List</code> contains at least
-%% one entry associated with <code>Key</code>, otherwise
-%% <code>false</code> is returned.
-
+-doc """
+Returns `true` if `List` contains at least one entry associated with `Key`,
+otherwise `false`.
+""".
 -spec is_defined(Key, List) -> boolean() when
       Key :: term(),
       List :: [term()].
@@ -189,8 +191,7 @@ is_defined(_Key, []) ->
 
 %% ---------------------------------------------------------------------
 
-%% @equiv get_value(Key, List, undefined)
-
+-doc(#{ equiv => get_value(Key, List, undefined) }).
 -spec get_value(Key, List) -> term() when
       Key :: term(),
       List :: [term()].
@@ -198,16 +199,13 @@ is_defined(_Key, []) ->
 get_value(Key, List) ->
     get_value(Key, List, undefined).
 
-%% @doc Returns the value of a simple key/value property in
-%% <code>List</code>. If <code>lookup(Key, List)</code> would yield
-%% <code>{Key, Value}</code>, this function returns the corresponding
-%% <code>Value</code>, otherwise <code>Default</code> is returned.
-%%
-%% @see lookup/2
-%% @see get_value/2
-%% @see get_all_values/2
-%% @see get_bool/2
+-doc """
+Returns the value of a simple key/value property in `List`. If
+[`lookup(Key, List)`](`lookup/2`) would yield `{Key, Value}`, this function
+returns the corresponding `Value`, otherwise `Default`.
 
+See also `get_all_values/2`, `get_bool/2`, `get_value/2`, `lookup/2`.
+""".
 -spec get_value(Key, List, Default) -> term() when
       Key :: term(),
       List :: [term()],
@@ -230,13 +228,10 @@ get_value(Key, [P | Ps], Default) ->
 get_value(_Key, [], Default) ->
     Default.
 
-%% @doc Similar to <code>get_value/2</code>, but returns the list of
-%% values for <em>all</em> entries <code>{Key, Value}</code> in
-%% <code>List</code>. If no such entry exists, the result is the empty
-%% list.
-%%
-%% @see get_value/2
-
+-doc """
+Similar to `get_value/2`, but returns the list of values for _all_ entries
+`{Key, Value}` in `List`. If no such entry exists, the result is the empty list.
+""".
 -spec get_all_values(Key, List) -> [term()] when
       Key :: term(),
       List :: [term()].
@@ -257,15 +252,23 @@ get_all_values(Key, [P | Ps]) ->
 get_all_values(_Key, []) ->
     [].
 
-%% @doc Similar to <code>get_all_values/2</code>, but each value is
-%% wrapped in a list unless it is already itself a list, and the
-%% resulting list of lists is concatenated. This is often useful for
-%% "incremental" options; e.g., <code>append_values(a, [{a, [1,2]}, {b,
-%% 0}, {a, 3}, {c, -1}, {a, [4]}])</code> will return the list
-%% <code>[1,2,3,4]</code>.
-%%
-%% @see get_all_values/2
+-doc """
+Similar to `get_all_values/2`, but each value is wrapped in a list unless it is
+already itself a list. The resulting list of lists is concatenated. This is
+often useful for "incremental" options.
 
+_Example:_
+
+```erlang
+append_values(a, [{a, [1,2]}, {b, 0}, {a, 3}, {c, -1}, {a, [4]}])
+```
+
+returns:
+
+```erlang
+[1,2,3,4]
+```
+""".
 -spec append_values(Key, ListIn) -> ListOut when
       Key :: term(),
       ListIn :: [term()],
@@ -292,14 +295,13 @@ append_values(_Key, []) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Returns the value of a boolean key/value option. If
-%% <code>lookup(Key, List)</code> would yield <code>{Key, true}</code>,
-%% this function returns <code>true</code>; otherwise <code>false</code>
-%% is returned.
-%%
-%% @see lookup/2
-%% @see get_value/2
+-doc """
+Returns the value of a boolean key/value option. If
+[`lookup(Key, List)`](`lookup/2`) would yield `{Key, true}`, this function
+returns `true`, otherwise `false`.
 
+See also `get_value/2`, `lookup/2`.
+""".
 -spec get_bool(Key, List) -> boolean() when
       Key :: term(),
       List :: [term()].
@@ -324,9 +326,7 @@ get_bool(_Key, []) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Returns an unordered list of the keys used in <code>List</code>,
-%% not containing duplicates.
-
+-doc "Returns an unordered list of the keys used in `List`, not containing duplicates.".
 -spec get_keys(List) -> [term()] when
       List :: [term()].
 
@@ -347,9 +347,7 @@ get_keys([], Keys) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Deletes all entries associated with <code>Key</code> from
-%% <code>List</code>.
-
+-doc "Deletes all entries associated with `Key` from `List`.".
 -spec delete(Key, List) -> List when
       Key :: term(),
       List :: [term()].
@@ -368,21 +366,19 @@ delete(_, []) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Substitutes keys of properties. For each entry in
-%% <code>ListIn</code>, if it is associated with some key <code>K1</code>
-%% such that <code>{K1, K2}</code> occurs in <code>Aliases</code>, the
-%% key of the entry is changed to <code>Key2</code>. If the same
-%% <code>K1</code> occurs more than once in <code>Aliases</code>, only
-%% the first occurrence is used.
-%%
-%% <p>Example: <code>substitute_aliases([{color, colour}], L)</code>
-%% will replace all tuples <code>{color, ...}</code> in <code>L</code>
-%% with <code>{colour, ...}</code>, and all atoms <code>color</code>
-%% with <code>colour</code>.</p>
-%%
-%% @see substitute_negations/2
-%% @see normalize/2
+-doc """
+Substitutes keys of properties. For each entry in `ListIn`, if it is associated
+with some key `K1` such that `{K1, K2}` occurs in `Aliases`, the key of the
+entry is changed to `K2`. If the same `K1` occurs more than once in `Aliases`,
+only the first occurrence is used.
 
+For example,
+[`substitute_aliases([{color, colour}], L)`](`substitute_aliases/2`) replaces
+all tuples `{color, ...}` in `L` with `{colour, ...}`, and all atoms `color`
+with `colour`.
+
+See also `normalize/2`, `substitute_negations/2`.
+""".
 -spec substitute_aliases(Aliases, ListIn) -> ListOut when
       Aliases :: [{Key, Key}],
       Key :: term(),
@@ -406,26 +402,24 @@ substitute_aliases_1([], P) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Substitutes keys of boolean-valued properties and simultaneously
-%% negates their values. For each entry in <code>ListIn</code>, if it is
-%% associated with some key <code>K1</code> such that <code>{K1,
-%% K2}</code> occurs in <code>Negations</code>, then if the entry was
-%% <code>{K1, true}</code> it will be replaced with <code>{K2,
-%% false}</code>, otherwise it will be replaced with <code>K2</code>,
-%% thus changing the name of the option and simultaneously
-%% negating the value given by <code>get_bool(ListIn)</code>. If the same
-%% <code>K1</code> occurs more than once in <code>Negations</code>, only
-%% the first occurrence is used.
-%%
-%% <p>Example: <code>substitute_negations([{no_foo, foo}], L)</code>
-%% will replace any atom <code>no_foo</code> or tuple <code>{no_foo,
-%% true}</code> in <code>L</code> with <code>{foo, false}</code>, and
-%% any other tuple <code>{no_foo, ...}</code> with <code>foo</code>.</p>
-%%
-%% @see get_bool/2
-%% @see substitute_aliases/2
-%% @see normalize/2
+-doc """
+Substitutes keys of boolean-valued properties and simultaneously negates their
+values.
 
+For each entry in `ListIn`, if it is associated with some key `K1` such
+that `{K1, K2}` occurs in `Negations`: if the entry was `{K1, true}`, it is
+replaced with `{K2, false}`, otherwise with `K2`, thus changing the name of the
+option and simultaneously negating the value specified by
+[`get_bool(Key, ListIn)`](`get_bool/2`). If the same `K1` occurs more than once
+in `Negations`, only the first occurrence is used.
+
+For example,
+[`substitute_negations([{no_foo, foo}], L)`](`substitute_negations/2`) replaces
+any atom `no_foo` or tuple `{no_foo, true}` in `L` with `{foo, false}`, and any
+other tuple `{no_foo, ...}` with `foo`.
+
+See also `get_bool/2`, `normalize/2`, `substitute_aliases/2`.
+""".
 -spec substitute_negations(Negations, ListIn) -> ListOut when
       Negations :: [{Key1, Key2}],
       Key1 :: term(),
@@ -460,40 +454,38 @@ substitute_negations_1([], P) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Expands particular properties to corresponding sets of
-%% properties (or other terms). For each pair <code>{Property,
-%% Expansion}</code> in <code>Expansions</code>, if <code>E</code> is
-%% the first entry in <code>ListIn</code> with the same key as
-%% <code>Property</code>, and <code>E</code> and <code>Property</code>
-%% have equivalent normal forms, then <code>E</code> is replaced with
-%% the terms in <code>Expansion</code>, and any following entries with
-%% the same key are deleted from <code>ListIn</code>.
-%%
-%% <p>For example, the following expressions all return <code>[fie, bar,
-%% baz, fum]</code>:
-%% <ul>
-%%   <li><code>expand([{foo, [bar, baz]}],
-%%                    [fie, foo, fum])</code></li>
-%%   <li><code>expand([{{foo, true}, [bar, baz]}],
-%%                    [fie, foo, fum])</code></li>
-%%   <li><code>expand([{{foo, false}, [bar, baz]}],
-%%                    [fie, {foo, false}, fum])</code></li>
-%% </ul>
-%% However, no expansion is done in the following call:
-%% <ul>
-%%   <li><code>expand([{{foo, true}, [bar, baz]}],
-%%                    [{foo, false}, fie, foo, fum])</code></li>
-%% </ul>
-%% because <code>{foo, false}</code> shadows <code>foo</code>.</p>
-%%
-%% <p>Note that if the original property term is to be preserved in the
-%% result when expanded, it must be included in the expansion list. The
-%% inserted terms are not expanded recursively. If
-%% <code>Expansions</code> contains more than one property with the same
-%% key, only the first occurrance is used.</p>
-%%
-%% @see normalize/2
+-doc """
+Expands particular properties to corresponding sets of properties (or other
+terms).
 
+For each pair `{Property, Expansion}` in `Expansions`: if `E` is the
+first entry in `ListIn` with the same key as `Property`, and `E` and `Property`
+have equivalent normal forms, then `E` is replaced with the terms in
+`Expansion`, and any following entries with the same key are deleted from
+`ListIn`.
+
+For example, the following expressions all return `[fie, bar, baz, fum]`:
+
+```erlang
+expand([{foo, [bar, baz]}], [fie, foo, fum])
+expand([{{foo, true}, [bar, baz]}], [fie, foo, fum])
+expand([{{foo, false}, [bar, baz]}], [fie, {foo, false}, fum])
+```
+
+However, no expansion is done in the following call because `{foo, false}`
+shadows `foo`:
+
+```erlang
+expand([{{foo, true}, [bar, baz]}], [{foo, false}, fie, foo, fum])
+```
+
+Notice that if the original property term is to be preserved in the result when
+expanded, it must be included in the expansion list. The inserted terms are not
+expanded recursively. If `Expansions` contains more than one property with the
+same key, only the first occurrence is used.
+
+See also `normalize/2`.
+""".
 -spec expand(Expansions, ListIn) -> ListOut when
       Expansions :: [{Property :: property(), Expansion :: [term()]}],
       ListIn :: [term()],
@@ -576,27 +568,27 @@ flatten([]) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Passes <code>List</code> through a sequence of
-%% substitution/expansion stages. For an <code>aliases</code> operation,
-%% the function <code>substitute_aliases/2</code> is applied using the
-%% given list of aliases; for a <code>negations</code> operation,
-%% <code>substitute_negations/2</code> is applied using the given
-%% negation list; for an <code>expand</code> operation, the function
-%% <code>expand/2</code> is applied using the given list of expansions.
-%% The final result is automatically compacted (cf.
-%% <code>compact/1</code>).
-%%
-%% <p>Typically you want to substitute negations first, then aliases,
-%% then perform one or more expansions (sometimes you want to pre-expand
-%% particular entries before doing the main expansion). You might want
-%% to substitute negations and/or aliases repeatedly, to allow such
-%% forms in the right-hand side of aliases and expansion lists.</p>
-%%
-%% @see substitute_aliases/2
-%% @see substitute_negations/2
-%% @see expand/2
-%% @see compact/1
+-doc """
+Passes `ListIn` through a sequence of substitution/expansion stages. For an
+`aliases` operation, function `substitute_aliases/2` is applied using the
+specified list of aliases:
 
+- For a `negations` operation,
+  [`substitute_negations/2`](`substitute_negations/2`) is applied using the
+  specified negation list.
+- For an `expand` operation, function `expand/2` is applied using the specified
+  list of expansions.
+
+The final result is automatically compacted (compare `compact/1`).
+
+Typically you want to substitute negations first, then aliases, then perform one
+or more expansions (sometimes you want to pre-expand particular entries before
+doing the main expansion). You might want to substitute negations and/or aliases
+repeatedly, to allow such forms in the right-hand side of aliases and expansion
+lists.
+
+See also `substitute_negations/2`.
+""".
 -spec normalize(ListIn, Stages) -> ListOut when
       ListIn :: [term()],
       Stages :: [Operation],
@@ -622,20 +614,26 @@ apply_stages(L, []) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Partitions <code>List</code> into a list of sublists and a
-%% remainder. <code>Lists</code> contains one sublist for each key in
-%% <code>Keys</code>, in the corresponding order. The relative order of
-%% the elements in each sublist is preserved from the original
-%% <code>List</code>. <code>Rest</code> contains the elements in
-%% <code>List</code> that are not associated with any of the given keys,
-%% also with their original relative order preserved.
-%%
-%% <p>Example:<pre>
-%% split([{c, 2}, {e, 1}, a, {c, 3, 4}, d, {b, 5}, b], [a, b, c])</pre>
-%% returns<pre>
-%% {[[a], [{b, 5}, b],[{c, 2}, {c, 3, 4}]], [{e, 1}, d]}</pre>
-%% </p>
+-doc """
+Partitions `List` into a list of sublists and a remainder.
 
+`Lists` contains one sublist for each key in `Keys`, in the corresponding order.
+The relative order of the elements in each sublist is preserved from the original `List`.
+`Rest` contains the elements in `List` that are not associated with any of the
+specified keys, also with their original relative order preserved.
+
+_Example:_
+
+```erlang
+split([{c, 2}, {e, 1}, a, {c, 3, 4}, d, {b, 5}, b], [a, b, c])
+```
+
+returns:
+
+```erlang
+{[[a], [{b, 5}, b],[{c, 2}, {c, 3, 4}]], [{e, 1}, d]}
+```
+""".
 -spec split(List, Keys) -> {Lists, Rest} when
       List :: [term()],
       Keys :: [term()],
@@ -675,25 +673,31 @@ maps_prepend(Key, Val, Dict) ->
 
 %% ---------------------------------------------------------------------
 
-%% @doc Converts the property list <code>List</code> to a map.
-%%
-%% Shorthand atom values in <code>List</code> will be expanded to an
-%% association of the form <code>Atom => true</code>. Tuples of the
-%% form <code>{Key, Value}</code> in <code>List</code> will be
-%% converted to an association of the form <code>Key => Value</code>.
-%% Anything else will be silently ignored.
-%%
-%% If the same key appears in <code>List</code> multiple times, the
-%% value of the one appearing nearest to the head of <code>List</code>
-%% will be in the result map, that is the value that would be returned
-%% by a call to <code>get_value/2</code> with this key.
-%%
-%% <p>Example:<pre>
-%% to_map([a, {b, 1}, {c, 2}, {c, 3}])</pre>
-%% returns<pre>
-%% #{a => true, b => 1, c => 2}</pre>
-%% </p>
+-doc """
+Converts the property list `List` to a map.
 
+Shorthand atom values in `List` will be expanded to an association of the form
+`Atom => true`. Tuples of the form `{Key, Value}` in `List` will be converted to
+an association of the form `Key => Value`. Anything else will be silently
+ignored.
+
+If the same key appears in `List` multiple times, the value of the one appearing
+nearest to the head of `List` will be in the result map, that is the value that
+would be returned by a call to [`get_value(Key, List)`](`get_value/2`).
+
+_Example:_
+
+```erlang
+to_map([a, {b, 1}, {c, 2}, {c, 3}])
+```
+
+returns:
+
+```erlang
+#{a => true, b => 1, c => 2}
+```
+""".
+-doc(#{since => <<"OTP 24.0">>}).
 -spec to_map(List) -> Map when
       List :: [Shorthand | {Key, Value} | term()],
       Map :: #{Shorthand => 'true', Key => Value},
@@ -719,12 +723,13 @@ to_map(List) ->
         List
     ).
 
-%% @doc Converts the property list <code>List</code> to a map
-%% after applying the normalizations given in <code>Stages</code>.
-%%
-%% @see normalize/2
-%% @see to_map/1
+-doc """
+Converts the property list `List` to a map after applying the normalizations
+given in `Stages`.
 
+See also `normalize/2`, `to_map/1`.
+""".
+-doc(#{since => <<"OTP 24.0">>}).
 -spec to_map(List, Stages) -> Map when
       List :: [term()],
       Stages :: [Operation],
@@ -739,8 +744,8 @@ to_map(List) ->
 to_map(List, Stages) ->
     to_map(apply_stages(List, Stages)).
 
-%% @doc Converts the map <code>Map</code> to a property list.
-
+-doc "Converts the map `Map` to a property list.".
+-doc(#{since => <<"OTP 24.0">>}).
 -spec from_map(Map) -> List when
     Map :: #{Key => Value},
     List :: [{Key, Value}],
